/*******************************************************************
  uLan Communication - uL_DRV - multiplatform uLan driver

  ul_c450.c	- chip driver for 16450

  (C) Copyright 1996-2004 by Pavel Pisa - project originator
        http://cmp.felk.cvut.cz/~pisa
  (C) Copyright 1996-2004 PiKRON Ltd.
        http://www.pikron.com
  (C) Copyright 2002-2004 Petr Smolik
  

  The uLan driver project can be used and distributed 
  in compliance with any of next licenses
   - GPL - GNU Public License
     See file COPYING for details.
   - LGPL - Lesser GNU Public License
   - MPL - Mozilla Public License
   - and other licenses added by project originator

  Code can be modified and re-distributed under any combination
  of the above listed licenses. If contributor does not agree with
  some of the licenses, he/she can delete appropriate line.
  WARNING: if you delete all lines, you are not allowed to
  distribute code or sources in any form.
 *******************************************************************/

/*******************************************************************/
/* Chip driver for 16450 */

#include "serial_reg.h"

#define u450_inb(base,port)	 ul_inb(base+port)
#define u450_outb(base,port,val) ul_outb(base+port,val)

#define U450_MCR_IE	0x0D	/* receive */
#define U450_MCR_OE	0x0E	/* transmit */
#define U450_MCR_OEQ	0x02	/* transmit ? */

#define U450_MSR_RxD	UART_MSR_CTS /* RxD line readback */
#define U450_MSR_TxD	UART_MSR_DSR /* TxD line readback */
#define U450_MSR_DTxD	UART_MSR_DDSR /* TxD line changed */

#define U450_LCR_UL	0x3B	/* receive/transmit */
#define U450_LCR_ULC	0x2B	/* transmit control character */
#define U450_LCR_ULB	0x7B	/* transmit line break */

#define U450_LSR_C	0x04	/* controll character received */
#define U450_LSR_ERR	0x1A	/* error in receiver */

/*** Test interrupt request state ***/
int u450_pool(ul_drv *udrv)
{
  unsigned u;
  u=u450_inb(udrv->port,UART_IIR);
  return (u&UART_IIR_NO_INT)?0:((u&UART_IIR_ID)|1);
};

/*** Wait end of transmit ***/
int u450_weot(ul_drv *udrv, int ret_code)
{
  int ret;
  unsigned u;
  uchar lsr_val;
  uchar uc;
  if(!udrv->chip_buff[0]||
     !(u450_inb(udrv->port,UART_MCR)&U450_MCR_OEQ)) 
  {
    UL_FRET;
    u450_outb(udrv->port,UART_LCR,U450_LCR_UL);
    return UL_RC_PROC;
  };
  u450_outb(udrv->port,UART_IER,UART_IER_RDI);
  lsr_val=u450_inb(udrv->port,UART_LSR);
  ret=UL_RC_WIRQ;
  if(!(~lsr_val&(UART_LSR_THRE|UART_LSR_TEMT)))
  {
    LOG_CHIO(" TMI");
    ret=UL_RC_PROC; /* was commented out */
  }
  if(lsr_val&U450_LSR_ERR) 
    ret=UL_RC_EFRAME;	/* frame error */
  if(lsr_val&UART_LSR_DR)
  {
    if((u=udrv->chip_buff[1])) udrv->chip_buff[1]=0;
     else if((u=udrv->chip_buff[0])) udrv->chip_buff[0]=0;
    uc=u450_inb(udrv->port,UART_RX);
    if((u^uc)&0xff||!u)
    {
      LOG_CHIO(" TE%c!",u?'0':'1');
      ret=UL_RC_EBADCHR;
    }else{
      LOG_CHIO(".");
      if(ret>=0&&!udrv->chip_buff[0])
	ret=UL_RC_PROC;
    }; 
  };
  if(ret!=UL_RC_WIRQ)
  {
    u450_outb(udrv->port,UART_LCR,U450_LCR_UL);
    UL_FRET;
  };
  LOG_CHIO(" weot ret %d ",ret);
  return ret;
};

/*** Switch from receive to transmit ***/
  int u450_sw_R2T_1(ul_drv *udrv, int ret_code);
  int u450_sw_R2T_2(ul_drv *udrv, int ret_code);
  int u450_sw_R2T_3(ul_drv *udrv, int ret_code);

int u450_sw_R2T(ul_drv *udrv, int ret_code)
{
  udrv->chip_buff[0]=udrv->chip_buff[1]=0;
 /* A_wait */
  UL_FCALL2(u450_weot,u450_sw_R2T_1);
  return UL_RC_PROC;
};

int u450_sw_R2T_1(ul_drv *udrv, int ret_code)
{
  if(ret_code<0) {UL_FRET;return ret_code;};
  u450_outb(udrv->port,UART_MCR,U450_MCR_IE);
  u450_outb(udrv->port,UART_IER,UART_IER_THRI);
  u450_outb(udrv->port,UART_TX,0);
  UL_FNEXT(u450_sw_R2T_2);
  if(u450_inb(udrv->port,UART_LSR)&UART_LSR_THRE) return UL_RC_PROC;
    else return UL_RC_WIRQ;
}; 

int u450_sw_R2T_2(ul_drv *udrv, int ret_code)
{
 /* A_wait1: */
  u450_inb(udrv->port,UART_MSR);
  u450_outb(udrv->port,UART_IER,UART_IER_MSI);
  UL_FNEXT(u450_sw_R2T_3);
  return UL_RC_WIRQ;
}; 

int u450_sw_R2T_3(ul_drv *udrv, int ret_code)
{
  uchar uc;
 #if 0
  static int reported_bad_chip=0;
 #endif
 /* A_wait2: */
  uc=u450_inb(udrv->port,UART_MSR);
  if((uc&(U450_MSR_DTxD|U450_MSR_TxD))!=U450_MSR_DTxD)
    return UL_RC_WIRQ;
 #if 0
  while(~u450_inb(udrv->port,UART_LSR)&
	(UART_LSR_TEMT|UART_LSR_THRE))
   if(!reported_bad_chip++)
     LOG_FATAL(KERN_CRIT "uLan u450_sw_R2T_3 : problematic IC => looping in ISR\n");
 #endif
  UL_FRET;
  return UL_RC_PROC;
}; 

/*** Receive character into char_buff ***/
  int u450_recch_1(ul_drv *udrv, int ret_code);
  int u450_recch_2(ul_drv *udrv, int ret_code);

int u450_recch(ul_drv *udrv, int ret_code)
{
  udrv->char_buff=0;
  if(u450_inb(udrv->port,UART_MCR)&U450_MCR_OEQ)
  {
    UL_FCALL2(u450_weot,u450_recch_1);
    return UL_RC_PROC;
  };
  UL_FNEXT(u450_recch_2);
  return UL_RC_PROC;
};
 
int u450_recch_1(ul_drv *udrv, int ret_code)
{
  u450_outb(udrv->port,UART_MCR,U450_MCR_IE);
  if(ret_code<0) {UL_FRET;return ret_code;};
  UL_FNEXT(u450_recch_2);
  return UL_RC_PROC;
};

INLINE int u450_recch_sub(ul_drv *udrv,int lsr_val,int rec_char)
{ /* helper function to process received character */
  if(lsr_val&U450_LSR_C) udrv->last_ctrl=rec_char|=0x100; 
   else udrv->last_ctrl&=0x7F; 
  udrv->char_buff=rec_char;
  udrv->xor_sum^=rec_char;udrv->xor_sum++;
  if(lsr_val&U450_LSR_ERR)
  {
    LOG_CHIO(" ER:%03X",udrv->char_buff);
    return UL_RC_EFRAME; /* frame error */
  };
  LOG_CHIO(" R:%03X",udrv->char_buff);
  return UL_RC_PROC;
};
 
int u450_recch_2(ul_drv *udrv, int ret_code)
{
  uchar lsr_val;
  u450_outb(udrv->port,UART_IER,UART_IER_RDI);
  lsr_val=u450_inb(udrv->port,UART_LSR);
  if(!(lsr_val&UART_LSR_DR)) return UL_RC_WIRQ;
  UL_FRET;
  return u450_recch_sub(udrv,lsr_val,u450_inb(udrv->port,UART_RX));
};

/*** Send character from char_buff ***/
  int u450_sndch_1(ul_drv *udrv, int ret_code);
  int u450_sndch_2(ul_drv *udrv, int ret_code);

int u450_sndch(ul_drv *udrv, int ret_code)
{
  uchar uc;
  uc=(udrv->char_buff&0x100)?U450_LCR_ULC:U450_LCR_UL;
  if(u450_inb(udrv->port,UART_LCR)!=uc)
  { /* change between control and data character format */
   if((u450_inb(udrv->port,UART_MCR)&U450_MCR_OEQ))
   { /* wait to end of transmit */
     UL_FCALL2(u450_weot,u450_sndch_1);
     return UL_RC_PROC;
   }else{ /* direction switch necessary */
     udrv->chip_buff[0]=udrv->chip_buff[1]=0;
     UL_FCALL2(u450_sw_R2T_1,u450_sndch_1);
     UL_FCALL2(u450_weot,u450_sw_R2T_1);
     return UL_RC_PROC;
   };
  };
  if(!(u450_inb(udrv->port,UART_MCR)&U450_MCR_OEQ))
  { /* direction switch necessary */
    udrv->chip_buff[0]=udrv->chip_buff[1]=0;
    UL_FCALL2(u450_sw_R2T_1,u450_sndch_1);
    UL_FCALL2(u450_weot,u450_sw_R2T_1);
  } else UL_FNEXT(u450_sndch_1);
  return UL_RC_PROC;
};

int u450_sndch_1(ul_drv *udrv, int ret_code)
{
  uchar uc;
  if(ret_code<0) {UL_FRET;return ret_code;};
  u450_outb(udrv->port,UART_MCR,U450_MCR_OE);
  if(udrv->char_buff&0x100)
  { /* store last ctrl for connect line busy/ready */
    udrv->last_ctrl=udrv->char_buff;
    uc=U450_LCR_ULC;
  }else{ 
    uc=U450_LCR_UL;
  };
  udrv->chip_buff[1]=udrv->chip_buff[0];
  udrv->chip_buff[0]=udrv->char_buff|0x1000; 
  udrv->xor_sum^=udrv->char_buff;udrv->xor_sum++;
  u450_outb(udrv->port,UART_IER,UART_IER_THRI);
  u450_outb(udrv->port,UART_LCR,uc);  /* control/data character */
  u450_outb(udrv->port,UART_TX,(uchar)udrv->char_buff);
  LOG_CHIO(" T:%03X",udrv->char_buff);
  UL_FNEXT(u450_sndch_2);
  return UL_RC_WIRQ;
}; 

int u450_sndch_2(ul_drv *udrv, int ret_code)
{
  uchar uc;
  unsigned u;
  uc=u450_inb(udrv->port,UART_LSR);
  UL_FRET;
  if(uc&U450_LSR_ERR)
      return UL_RC_EFRAME;	/* frame error */
  if(!(uc&UART_LSR_DR))
      return UL_RC_PROC;		/* no char */ 
  if((u=udrv->chip_buff[1])) udrv->chip_buff[1]=0;
    else if((u=udrv->chip_buff[0])) udrv->chip_buff[0]=0;
      else return UL_RC_PROC;
  uc=u450_inb(udrv->port,UART_RX);
  if((u^uc)&0xff||!u)
  {
    LOG_CHIO(" TE%c!",u?'2':'3');
    return UL_RC_EBADCHR;
  };
  LOG_CHIO(".");
  return UL_RC_PROC;
};
 
/*** Wait for time or received character ***/
  int u450_wait_1(ul_drv *udrv, int ret_code);
  int u450_wait_2(ul_drv *udrv, int ret_code);

int u450_wait(ul_drv *udrv, int ret_code)
{
  udrv->char_buff=0;
  if(u450_inb(udrv->port,UART_MCR)&U450_MCR_OEQ)
  {
    UL_FCALL2(u450_weot,u450_wait_1);
  } else UL_FNEXT(u450_wait_1);
  return UL_RC_PROC;
};

int u450_wait_1(ul_drv *udrv, int ret_code)
{
 uchar lsr_val;
 if(ret_code<0) {UL_FRET;return ret_code;};
 u450_outb(udrv->port,UART_MCR,U450_MCR_IE);
 u450_outb(udrv->port,UART_IER,UART_IER_THRI|UART_IER_RDI);
 if((lsr_val=u450_inb(udrv->port,UART_LSR))&UART_LSR_DR)
 {
   UL_FRET;
   return u450_recch_sub(udrv,lsr_val,u450_inb(udrv->port,UART_RX));
 };
 if(!(lsr_val&UART_LSR_THRE)) return UL_RC_WIRQ;
 if(!udrv->wait_time)
 {
   UL_FRET;
   LOG_CHIO(" Timeout!");
   return UL_RC_ETIMEOUT;
 };
 UL_FNEXT(u450_wait_2);
 u450_outb(udrv->port,UART_TX,0);
 return UL_RC_WIRQ;
};

int u450_wait_2(ul_drv *udrv, int ret_code)
{
  uchar lsr_val;
  if((lsr_val=u450_inb(udrv->port,UART_LSR))&UART_LSR_DR)
  {
    UL_FRET;
    return u450_recch_sub(udrv,lsr_val,u450_inb(udrv->port,UART_RX));
  };
  if(!udrv->wait_time--)
  {
    UL_FRET;
    LOG_CHIO(" Timeout!");
    return UL_RC_ETIMEOUT;
  };
  u450_outb(udrv->port,UART_TX,0);
  return UL_RC_WIRQ;
};

/*** Connect to RS485 bus ***/
  int u450_connect_1(ul_drv *udrv, int ret_code);
  int u450_connect_2(ul_drv *udrv, int ret_code);

int u450_connect(ul_drv *udrv, int ret_code)
{
  unsigned u;
  LOG_CHIO(" C_0 ");
  u=udrv->last_ctrl;
  udrv->wait_time=((udrv->my_adr-u-1)&0xF)+4; 
  if(((u&0x180)!=0x180)||
     (u==0x1FF)) udrv->wait_time+=0x10;
  udrv->last_ctrl=0;
  udrv->char_buff=0;
  udrv->chip_temp=(udrv->my_adr&0x3F)|0x40;
  u450_inb(udrv->port,UART_MSR);
  UL_FCALL2(*udrv->fnc_wait,u450_connect_1);
  return UL_RC_PROC;
};

int u450_connect_1(ul_drv *udrv, int ret_code)
{
  LOG_CHIO(" C_1 ");
  if(ret_code!=UL_RC_ETIMEOUT)
  { UL_FRET;
    LOG_CHIO(" EARBIT1!");
    return UL_RC_EARBIT;
  };
  /* check for RxD low */
  if(u450_inb(udrv->port,UART_MSR)&U450_MSR_RxD)
  { UL_FRET;
    LOG_CHIO(" EARBIT2!");
    return UL_RC_EARBIT;
  };
  u450_outb(udrv->port,UART_LCR,U450_LCR_ULB);
  u450_inb(udrv->port,UART_LSR);
  u450_outb(udrv->port,UART_IER, UART_IER_RLSI);
  u450_outb(udrv->port,UART_MCR,U450_MCR_OE);
  UL_FNEXT(u450_connect_2);
  /* wait for break condition now */
  return UL_RC_WIRQ;
};

int u450_connect_2(ul_drv *udrv, int ret_code)
{
  LOG_CHIO(" C_2 ");
  if(!u450_inb(udrv->port,UART_LSR)&UART_LSR_BI)
    return UL_RC_WIRQ;
  u450_outb(udrv->port,UART_LCR,U450_LCR_ULC);
  u450_outb(udrv->port,UART_MCR,U450_MCR_IE);
  u450_inb(udrv->port,UART_LSR);
  u450_inb(udrv->port,UART_RX);
  if(udrv->chip_temp==1)
  {LOG_CHIO(" Connected");
    UL_FRET;
  }else{ 
    udrv->wait_time=(udrv->chip_temp&3)+1; 
    udrv->chip_temp>>=2;
    UL_FCALL2(*udrv->fnc_wait,u450_connect_1);
  };
  return UL_RC_PROC;
};

/*
 UL_FNEXT(u450_recch);
 UL_FCALL(u450_recch);
 UL_FCALL2(u450_recch,u450_recch);
 UL_FRET;
*/

static char *u450_port_name="ulan_u450";

/*** 16450 initialize ports ***/
int u450_pinit(ul_drv *udrv)
{
  unsigned u;
  int baud=udrv->baud_val;
  if (UL_REQ_IOS(udrv->port,8,u450_port_name)<0) 
  { LOG_FATAL(KERN_CRIT "uLan u450_pinit : cannot reguest ports !\n");
    return UL_RC_EPORT;
  };
 
  if (!baud) baud=19200;
  u=(0x1C200+baud/2)/baud;
  if (u>0xFFFF) u=0xFFFF;
  udrv->baud_div=u;

  udrv->chip_buff[0]=udrv->chip_buff[1]=0;
  u450_outb(udrv->port,UART_MCR,UART_MCR_OUT1|UART_MCR_DTR);
  u450_outb(udrv->port,UART_LCR,UART_LCR_DLAB);
  u450_outb(udrv->port,(uchar)UART_DLL,(uchar)udrv->baud_div);
  u450_outb(udrv->port,(uchar)UART_DLM,(uchar)(udrv->baud_div>>8));
  u450_outb(udrv->port,UART_LCR,U450_LCR_UL);
  u450_outb(udrv->port,UART_IER,0);
  u450_outb(udrv->port,UART_FCR,0);	/* disable FIFOs */
  u450_inb(udrv->port,UART_LSR);		/* reset errors */
  u450_inb(udrv->port,UART_RX);		/* and other irq */
  u450_outb(udrv->port,UART_MCR,U450_MCR_IE); /* connect irq */
  return UL_RC_PROC;
};

/*** 16450 deinitialize ports ***/
int u450_pdone(ul_drv *udrv)
{
  u450_outb(udrv->port,UART_IER, 0);	/* disable interrupts */
  u450_outb(udrv->port,UART_MCR, U450_MCR_IE&3);  /* transmitter off */
  u450_inb(udrv->port,UART_IIR);		/* flush irq */
  u450_inb(udrv->port,UART_LSR);		/* flush irq */
  UL_REL_IOS(udrv->port,8);
  return UL_RC_PROC;
};

/*** 16450 generate irq for irq_probe */
int u450_genirq(ul_drv *udrv,int param)
{
  if(param) 
  {
    u450_outb(udrv->port,UART_MCR, U450_MCR_IE);  /* transmitter off */
    u450_outb(udrv->port,UART_IER, UART_IER_MSI);	/* enable interrupts */
    u450_outb(udrv->port,UART_LCR, U450_LCR_ULB);	/* trig break */
  }else{
    u450_outb(udrv->port,UART_IER, 0);		/* disable interrupts */
    u450_outb(udrv->port,UART_MCR, U450_MCR_IE);	/* transmitter off */
    u450_inb(udrv->port,UART_MSR);		/* flush irq */
  };
  return UL_RC_PROC;
};

/* support for hardware tests */
int u450_hwtest(ul_drv *udrv,int param)
{
  unsigned u;
  switch(param)
  {
   case 0x10:
   case 0x11:
	u450_outb(udrv->port,UART_IER, 0); /* disable interrupts */
	u450_outb(udrv->port,UART_MCR, U450_MCR_OE); /* transmitter on */
	u450_outb(udrv->port,UART_LCR, param&1?
	             U450_LCR_UL:U450_LCR_ULB); /* set TD lines */
   case 0x12:
        u=u450_inb(udrv->port,UART_MSR);
	return u450_inb(udrv->port,UART_LSR)|(u<<8)|
	                (u&(U450_MSR_TxD)?0:0x100000)|
			(u&(U450_MSR_RxD)?0:0x10000);
   case 0x13:
	u450_outb(udrv->port,UART_MCR, U450_MCR_IE); /* transmitter off */
        u=u450_inb(udrv->port,UART_MSR);
	return u450_inb(udrv->port,UART_LSR)|(u<<8)|
	                (u&(U450_MSR_TxD)?0:0x100000)|
			(u&(U450_MSR_RxD)?0:0x10000);
  }
  return UL_RC_ENOFNC;
};

/*** Control functions of chip driver  ***/
int u450_cctrl(ul_drv *udrv, int ctrl_fnc, int param)
{
  switch (ctrl_fnc)
  {
   case UL_CC_DIN:	/* switch off line transmitter */
	{ u450_outb(udrv->port,UART_MCR, U450_MCR_IE); 
	  u450_outb(udrv->port,UART_LCR, U450_LCR_UL); 
	  return UL_RC_PROC;
	};
   case UL_CC_DOUT:	/* switch on line transmitter */
	{ u450_outb(udrv->port,UART_MCR, U450_MCR_OE);
	  u450_outb(udrv->port,UART_LCR, U450_LCR_UL); 
	  return UL_RC_PROC;
	};
   case UL_CC_PINIT:	/* initialize ports (parameter baud)*/
  	{ return u450_pinit(udrv);
  	};
   case UL_CC_PDONE:	/* deinitialize ports */
  	{ return u450_pdone(udrv);
  	};
   case UL_CC_GENIRQ:	/* generate irq for irq_probe */
  	{return u450_genirq(udrv,param);
  	};
   case UL_CC_HWTEST:	/* support for hardware testing */
  	{return u450_hwtest(udrv,param);
  	};
  };
  return UL_RC_ENOFNC;
};

/*** 16450 chip driver initialize ***/
int u450_init(ul_drv *udrv, int port, int irq, int baud, int options)
{
  unsigned u;
  unsigned old_mcr;
  unsigned old_lcr;
  if (UL_REQ_IOS(port,8,u450_port_name)<0)
  { LOG_FATAL(KERN_CRIT "uLan u450_init : cannot reguest ports !\n");
    return UL_RC_EPORT;
  };
  
  LOG_PORTS("uLan u450_init : probe MCR\n");
  old_mcr=u450_inb(port,UART_MCR);
  LOG_PORTS("uLan u450_init : MCR = 0x%02X\n",old_mcr);

  u450_outb(port,UART_MCR,UART_MCR_OUT1|UART_MCR_DTR);
  #ifdef FOR_LINUX_KERNEL
   set_current_state(TASK_INTERRUPTIBLE);
   schedule_timeout(ULD_HZ/10);
  #elif defined(_WIN32)
   /* !!!!!!!!!!!!!! */
   /* KeStallExecutionProcessor(20000); */
   { LARGE_INTEGER delay;
     delay=RtlConvertLongToLargeInteger(-30*10000);
     KeDelayExecutionThread(KernelMode,FALSE,&delay);
   }
  #else
   sleep(1);
  #endif
  old_lcr=u450_inb(port,UART_LCR);
  u450_outb(port,UART_LCR,UART_LCR_SBC);
  #ifdef FOR_LINUX_KERNEL
   set_current_state(TASK_INTERRUPTIBLE);
   schedule_timeout(ULD_HZ/10);
  #elif defined(_WIN32)
   /* !!!!!!!!!!!!!! */
   /* KeStallExecutionProcessor(2000); */
   { LARGE_INTEGER delay;
     delay=RtlConvertLongToLargeInteger(-10*10000);
     KeDelayExecutionThread(KernelMode,FALSE,&delay);
   }
  #else
   sleep(1);
  #endif
  if(u450_inb(port,UART_MSR)&U450_MSR_TxD)
  { u450_outb(port,UART_LCR,U450_LCR_UL);
    LOG_PORTS("uLan u450_init : MCR test 1 OK\n");
    #ifdef FOR_LINUX_KERNEL
     set_current_state(TASK_INTERRUPTIBLE);
     schedule_timeout(ULD_HZ/10);
    #elif defined(_WIN32)
     /* !!!!!!!!!!!!!! */
     /* KeStallExecutionProcessor(2000); */
     { LARGE_INTEGER delay;
       delay=RtlConvertLongToLargeInteger(-10*10000);
       KeDelayExecutionThread(KernelMode,FALSE,&delay);
     }
    #else
     sleep(1);
    #endif
    if(!(u450_inb(port,UART_MSR)&U450_MSR_TxD))
    { LOG_PORTS("uLan u450_init : MCR test 2 OK\n");
      udrv->chip_options=options;
      if(!baud) baud=19200;
      udrv->baud_val=baud;
      u=(0x1C200+baud/2)/baud;
      if (u>0xFFFF) u=0xFFFF;
      udrv->baud_div=u;
      udrv->port=port;
      udrv->irq=irq; 
      udrv->fnc_recch=&u450_recch;
      udrv->fnc_sndch=&u450_sndch;
      udrv->fnc_wait=&u450_wait;
      udrv->fnc_pool=&u450_pool;
      udrv->fnc_connect=&u450_connect;
      udrv->fnc_cctrl=&u450_cctrl;
      UL_REL_IOS(port,8);
      return 0;
    };
    LOG_PORTS("uLan u450_init : MCR test 2 Error\n");
  };
  u450_outb(port,UART_LCR,(uchar)old_lcr);
  u450_outb(port,UART_MCR,(uchar)old_mcr);
  LOG_PORTS("uLan u450_init : MCR test Error\n");
  UL_REL_IOS(port,8);
  return UL_RC_EPORT;
};

